# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, BoxingFactory
from yay.errors import NoMatching

class Mapping(Node):

    """
    An unordered collection of key/value pairs
    """

    def __init__(self, predecessor):
        self.predecessor = predecessor
        if predecessor:
            predecessor.set_parent(self)
        self.value = {}

    def set(self, key, val):
        #val.chain = self.value.get("key", None)
        self.value[key] = val
        val.set_parent(self)

    def get(self, key):
        if key in self.value:
            return self.value[key]
        if self.predecessor:
            return self.predecessor.get(key)
        self.error(NoMatching("Not found: '%s'" % key))

    def keys(self):
        keys = set(self.value.keys())
        if self.predecessor:
            keys.update(self.predecessor.expand().keys())
        return list(keys)

    def resolve(self):
        data = {}
        for key in self.keys():
            try:
                key = key.encode('ascii')
            except UnicodeEncodeError:
                pass

            data[key] = self.get(key).resolve()
        return data

    def walk(self):
        for itm in self.value.items():
            yield itm
        yield self.predecessor

    def clone(self):
        p = None
        if self.predecessor:
            p = self.predecessor.clone()

        m = Mapping(p)
        for k, v in self.value.items():
            m.set(k, v.clone())
        return m

    def __iter__(self):
        return iter(BoxingFactory.box(v) for v in sorted(self.keys()))


def box_dict(val):
    m = Mapping(None)
    for k, v in val.items():
        m.set(k, BoxingFactory.box(v))
    return m
BoxingFactory.register(lambda x: isinstance(x, dict), box_dict)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaml.reader import Reader
from yaml.scanner import Scanner
from yaml.parser import Parser

from yay.openers import Openers
from yay.composer import Composer

class Loader(Reader, Scanner, Parser, Composer):

    @property
    def special_term(self):
        return self.parent.special_term

    @property
    def openers(self):
        return self.parent.openers

    def __init__(self, stream, name="<Unknown>", parent=None, secret=False):
        self.name = name
        self.parent = parent

        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self, secret=secret)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.pyparsing import *

from yay import nodes

class Parser(object):
    def __init__(self, composer):
        self.composer = composer

        if composer and composer.secret:
            self.secret = True
        else:
            self.secret = False

        self.setup_parser()

    def box(self, value):
        b = nodes.Boxed(value)
        b.secret = self.secret
        return b

    def dollar(self, str, words, tokens):
        return self.box("$")

    def boxed_string(self, str, words, tokens):
        return self.box(tokens[0])

    def boxed_int(self, str, words, tokens):
        return self.box(int(tokens[0]))

    def boxed_octal(self, str, words, tokens):
        return self.box(int(tokens[0], 8))

    def concatenation(self, str, words, tokens):
        if len(tokens) == 1:
            return tokens[0]
        c = nodes.Concatenation(*tokens)
        c.secret = self.secret
        return c

    def function_call_action(self, s, w, t):
        return nodes.Function(t[0], t[1])

    def filter_bin_comparison_action(self, s, w, t):
        cls = {
            "=": nodes.Equal,
            "!=": nodes.NotEqual,
            "<": nodes.LessThan,
            "<=": nodes.LessThanEqual,
            ">": nodes.GreaterThan,
            ">=": nodes.GreaterThanEqual,
            }[t[0][1]]
        return cls(t[0][0], t[0][2])

    def filter_expression_action(self, s, w, t):
        node = t[0]
        for i in range(1, len(t)):
            if t[i] == "and":
                node = nodes.And(node, t[i+1])
            elif t[i] == "or":
                node = nodes.Or(node, t[i+1])
            i += 1
        return node

    def handle_expression(self, s, w, t):
        if len(t[0]) == 1:
            return t[0]

        t = t[0]

        node = nodes.Else(t[0])

        for i in range(1, len(t)):
            if i % 2:
                if t[i] != "else":
                    #FIXME: Raise some kind of parasing error
                    pass
            else:
                node.append(t[i])

        return [node]

    def index_access_action(self, s, w, t):
        return nodes.Access(None, t[0])

    def full_expression_action(self, s, w, t):
        node = None
        for token in t:
            if not isinstance(token, nodes.Node):
                node = nodes.Access(node, nodes.Boxed(token))
            else:
                token.container = node
                if node:
                    node.set_parent(token)
                node = token

        return node

    def ugh(self, s, w, t):
        if not t or not t[0]:
            return []
        return self.boxed_string(s, w, t)

    def inline_call(self, s, w, t):
        return nodes.Call(self.composer, t[0])

    def setup_parser(self):
        ELSE = Keyword("else")
        AND = Keyword("and")
        OR = Keyword("or")
        IN = Keyword("in")
        BINOP = oneOf("= != < > <= >=")

        identifier = Word(alphanums+"_") | Keyword("@")
        arithSign = Word("+-",exact=1)

        octNum = Combine(Optional(arithSign) + Suppress("0") + Word(nums)).setParseAction(self.boxed_octal)
        intNum = Combine(Optional(arithSign) + Word(nums)).setParseAction(self.boxed_int)

        expression = Forward()

        macro_call = Word(alphanums+"_.") + Suppress("!")
        macro_call.setParseAction(self.inline_call)

        function_identifier = Word(alphanums+"_")
        function_call = function_identifier + Group(Suppress("(") + Optional(expression + ZeroOrMore(Suppress(",") + expression)) + Suppress(")"))
        function_call.setParseAction(self.function_call_action)

        filterExpression = Forward()

        filter_bin_comparison = Group(expression + BINOP + expression)
        filter_bin_comparison.setParseAction(self.filter_bin_comparison_action)

        filterCondition = (
            filter_bin_comparison |
            ( "(" + filterExpression + ")" )
        )

        filterExpression << filterCondition + ZeroOrMore((AND|OR) + filterExpression)
        filterExpression.setParseAction(self.filter_expression_action)

        full_list_access = Suppress("[") + filterExpression + Suppress("]")
        full_list_access.setParseAction(lambda s, w, t: nodes.Filter(None, t[0]))

        listAccess = Suppress("[") + expression + Suppress("]")
        listAccess.setParseAction(self.index_access_action)


        fullExpression = identifier + ZeroOrMore(
            full_list_access |
            listAccess |
            Suppress(".") + identifier
            )
        fullExpression.setParseAction(self.full_expression_action)

        expression_part = (
            octNum |
            intNum |
            macro_call |
            function_call |
            fullExpression
            )

        expression << Group(expression_part + Optional(ELSE + expression)).setParseAction(self.handle_expression)

        bracketed_expression = Suppress("${").leaveWhitespace() + expression + Suppress("}").leaveWhitespace()

        myrol = restOfLine.copy().setParseAction(self.ugh)

        templated_string = ZeroOrMore(
            Keyword("$$").setParseAction(self.dollar) |
            bracketed_expression |
            SkipTo("${").leaveWhitespace().setParseAction(self.boxed_string)
            ) + myrol
        templated_string.setParseAction(self.concatenation)

        foreachif = Optional(Keyword("if") + filterExpression)

        foreach_statement = identifier + Suppress("in") + expression + Optional(Keyword("chain") | Keyword("nochain") | Keyword("flatten")) + foreachif
        as_statement = expression + Suppress("as") + identifier

        self.templated_string = templated_string
        self.foreach_statement = foreach_statement
        self.as_statement = as_statement
        self.expression = expression




from django.db import models


class Car(models.Model):
    name = models.CharField(max_length=256, primary_key=True)

    @classmethod
    def some_callable(cls):
        return list(Part.objects.all())

    @classmethod
    def some_generator(cls):
        for p in Part.objects.all():
            yield p

class Part(models.Model):
    name = models.CharField(max_length=256, primary_key=True)
    car = models.ForeignKey(Car, related_name='parts')




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class StringPart(object):
    """ I represent a portion of a string which may or may not be sensitive """

    __slots__ = ("value", "secret", )

    def __init__(self, value, secret):
        self.value = value
        self.secret = secret

    @property
    def protected(self):
        if self.secret:
            return "*" * 5
        return self.value

    @property
    def unprotected(self):
        return self.value

    def __getstate__(self):
        return dict(value=self.value, secret=self.secret)

    def __setstate__(self, state):
        self.value = state['value']
        self.secret = state['secret']


class String(object):
    """ I represent a string which contains sensitive parts """

    def __init__(self, parts=None):
        self.parts = []
        if parts:
            self.extend(parts)

    def add(self, value):
        if isinstance(value, String):
            self.parts.extend(value.parts)
        else:
            self.parts.append(StringPart(value, False))

    def add_secret(self, value):
        self.parts.append(StringPart(value, True))

    def extend(self, value):
        [self.add(v) for v in value]

    def as_list(self, secret=True):
        """
        Return the string parts as a list

        If secret is True (the default) then the sensitive parts of the string
        will be obfuscated. If it is False, they will not.
        """
        if secret:
            return [unicode(p.protected) for p in self.parts]
        return [unicode(p.unprotected) for p in self.parts]

    def __str__(self):
        return "".join([str(p.protected) for p in self.parts])

    def __unicode__(self):
        return u"".join([unicode(p.protected) for p in self.parts])

    @property
    def protected(self):
        return u"".join([unicode(p.protected) for p in self.parts])

    @property
    def unprotected(self):
        return u"".join([unicode(p.unprotected) for p in self.parts])




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from yay import String


class TestString(unittest.TestCase):

    def test_normal_string(self):
        a = String(["Just a normal string"])

        self.failUnlessEqual(unicode(a), "Just a normal string")
        self.failUnlessEqual(a.unprotected, "Just a normal string")

        self.failUnlessEqual(a.as_list(secret=True), ["Just a normal string"])
        self.failUnlessEqual(a.as_list(secret=False), ["Just a normal string"])

    def test_partially_secret(self):
        a = "adduser -u fred -p "
        b = String()
        b.add_secret("password")
        c = " -h /home/fred"
        d = String([a,b,c])

        self.failUnlessEqual(unicode(d), "adduser -u fred -p ***** -h /home/fred")
        self.failUnlessEqual(d.unprotected, "adduser -u fred -p password -h /home/fred")

        self.failUnlessEqual(d.as_list(secret=True), ["adduser -u fred -p ", "*****", " -h /home/fred"])
        self.failUnlessEqual(d.as_list(secret=False), ["adduser -u fred -p ", "password", " -h /home/fred"])





from django.conf import settings
settings.configure(
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': ':memory:',
            },
        },
    INSTALLED_APPS=['yay.tests.dj'],
    TIME_ZONE = 'Europe/London',
    )

from unittest import TestCase

from yay.config import Config
from yay.tests.dj.models import Car, Part
from yay.nodes import BoxingFactory
from yay.nodes.datastore import dj
from django.core.management.commands.syncdb import Command as syncdb


class TestDjangoBase(TestCase):

    def setUp(self):
        syncdb().handle_noargs()

        c1 = Car(name="fred")
        c1.save()
        c2 = Car(name="bobby")
        c2.save()

        p1 = Part(name="wheel", car=c1)
        p1.save()
        p2 = Part(name="brake", car=c1)
        p2.save()
        p3 = Part(name="badger", car=c2)
        p3.save()


class TestDjangoBoxing(TestDjangoBase):

    def test_box_instance(self):
        b = BoxingFactory.box(Car(name="pudding"))
        self.assert_(isinstance(b, dj.Instance))

    def test_box_query(self):
        b = BoxingFactory.box(Car.objects.filter(name="pudding"))
        self.assert_(isinstance(b, dj.InstanceList))

    def test_box_manager(self):
        b = BoxingFactory.box(Car.objects)
        self.assert_(isinstance(b, dj.InstanceList))


class TestDjangoYay(TestDjangoBase):

    def setUp(self):
        TestDjangoBase.setUp(self)

        self.config = Config()
        self.config.load("""
            metadata.bind:
                type: djangostore
                model: yay.tests.dj.models
            """)

    def test_all_cars(self):
        self.config.load("""test: ${metadata.Car}""")
        l = self.config.get()["test"]

        self.failUnless(isinstance(l[0], dict))
        self.failUnlessEqual(l[0]["name"], "fred")

    def test_foreach_car(self):
        self.config.load("""
            test.foreach c in metadata.Car: ${c.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ["fred", "bobby"])

    def test_foreach_car_parts(self):
        self.config.load("""
            test.foreach c in metadata.Car:
              .foreach p in c.parts: ${p.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ['wheel', 'brake', 'badger'])

    def test_foreach_car_parts_if(self):
        self.config.load("""
            x: badger
            test.foreach c in metadata.Car:
              .foreach p in c.parts if p.name=x: ${p.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ["badger"])

    def test_foreach_car_parts_if_failing(self):
        self.config.load("""
            x: badge
            test.foreach c in metadata.Car:
              .foreach p in c.parts if p.name=x: ${p.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], [])

    def test_model_callable(self):
        self.config.load("""
            test.foreach p in metadata.Car.some_callable: ${p.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ['wheel','brake','badger'])

    def test_model_callable_generator(self):
        self.config.load("""
            test.foreach p in metadata.Car.some_generator: ${p.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ['wheel','brake','badger'])




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node
from yay.errors import ProgrammingError

class DataStoreType(type):

    stores = {}

    def __new__(meta, class_name, bases, new_attrs):
        cls = type.__new__(meta, class_name, bases, new_attrs)
        name = new_attrs.get('__name__', class_name.lower())
        if name in meta.stores:
            raise ProgrammingError("Redefinition of DataStoreType '%s'" % name)
        meta.stores[name] = cls
        return cls


class DataStore(object):
    __metaclass__ = DataStoreType


class Bind(Node):

    """
    Configure access to an external data store such as a database for
    access by Yay.

    This node is added to the graph when the parser encounters a ``bind`` function::

        mydb.bind:
            type: somebackend
            arg1: foo
            arg2: bar

    Bind finds an implementation of :py:class:`~yay.nodes.datastore.bind.DataStoreType`
    that has a name of ``somebackend`` and passes all other configuration to it.
    """

    def __init__(self, config):
        self.config = config
        self.backend = None

    def get_backend(self):
        if not self.backend:
            backend_type = self.config.get("type").resolve()
            self.backend = DataStoreType.stores[backend_type](self.config)
        return self.backend

    def get(self, key):
        """
        If a node tries to traverse inside this node we ask the backend for that key
        """
        return self.get_backend().get(key)

    def resolve(self):
        """
        If you try to resolve this node you get an empty mapping. Ideally you would
        get your expanded DataStore, but it could be a massive database so we decided
        not to bother.
        """
        return {}





# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node

class With(Node):

    def __init__(self, value, expression, label):
        super(With, self).__init__(value)
        self.expression = expression
        expression.set_parent(self)
        self.label = label

    def get_context(self, key):
        if key == self.label:
            return self.expression
        return super(With, self).get_context(key)

    def expand(self):
        return self.value.expand()

    def resolve(self):
        return self.value.resolve()

    def __repr__(self):
        return "With(%s)" % self.value

    def clone(self):
        return With(self.value.clone(), self.label_)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from yay.nodes import Node
from yay import String

class Unboxable(Exception):
    pass


class BoxingFactory(object):

    """ Sometimes we need to box complicated types that we might have special
    nodes for wrapping. The BoxingFactory will wrap any special nodes in the
    right wrapper type before falling back to the Boxed() wrapping method. """

    boxers = []

    @classmethod
    def register(cls, checker, boxer):
        cls.boxers.append((checker, boxer))

    @classmethod
    def box(cls, value):
        for checker, boxer in cls.boxers:
            if checker(value):
                return boxer(value)
        return Boxed(value)


def callable_boxer(val):
    """ Attempt to box callables """
    try:
        #inspect.getcallargs(val)
        return BoxingFactory.box(val())
    except TypeError:
        raise Unboxable("Cannot invoke callable that requires arguments")
BoxingFactory.register(callable, callable_boxer)


class Boxed(Node):
    """
    A Boxed node is some sort of literal and uninteresting value that
    is wrapped simply so we can put it in the graph

    """

    def __init__(self, value):
        self.value = value

    secret = False

    def resolve(self):
        """
        Boxed will do some simple and automatic processing of data when it
        resolves.

          * If it isn't a subclass of basestring it will be returned as is
          * `yes`, `true` and `on` will become python True
          * `no`, `false` and `off` will become python False
          * If it can be cast to an ``int`` it will be.
          * If it can be represented in ascii it will be
          * Otherwise a unicode string is returned
        """
        value = self.value

        if not isinstance(value, basestring):
            return value

        if self.secret:
            p = String()
            p.add_secret(value)
            return p

        if value.lower() in ("yes", "true", "on"):
            return True
        if value.lower() in ("no", "false", "off"):
            return False

        try:
            return int(value)
        except ValueError:
            pass

        #FIXME: could still be a float or a timestamp

        try:
            return value.encode('ascii')
        except UnicodeEncodeError:
            return value

    def get(self, key):
        """
        If the boxed value implements a list or dict like value access mechanism
        a Boxed node will allow those inner values to be boxed on demand so Yay
        can access them
        """
        if isinstance(self.value, list):
            return BoxingFactory.box(self.value[key])

        if hasattr(self.value, "get"):
            return BoxingFactory.box(self.value.get(key))

        if hasattr(self.value, key):
            return BoxingFactory.box(getattr(self.value, key))

        self.error("Unable to find '%s'" % key)

    def __repr__(self):
        return "Boxed(%s)" % self.value

    def clone(self):
        return Boxed(self.value)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from StringIO import StringIO

try:
    from sqlalchemy import Column, ForeignKey, String, Integer, create_engine
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import sessionmaker, relationship
    from sqlalchemy.orm.collections import InstrumentedList
    has_sqlalchemy = True
except ImportError:
    has_sqlalchemy = False

import yay
from yay.nodes import Node, Boxed, Sequence
from yay.nodes.datastore.bind import DataStore

class InstanceList(Node):

    """
    I am a list of Instance nodes - i.e. a one-to-many relationship
    """

    def __init__(self, values):
        self.values = values

    def get(self, idx):
        v = self.values[idx]

        if hasattr(v, "__tablename__"):
            return Instance(v)

        return Boxed(self.values[idx])

    def resolve(self):
        return list(self.__iter__())

    def __iter__(self):
        for i in range(len(self.values)):
            yield self.get(i)


class Instance(Node):

    def __init__(self, value):
        self.value = value

    def get(self, key):
        v = getattr(self.value, key)

        if hasattr(v, "__tablename__"):
            return Instance(v)

        if isinstance(v, InstrumentedList):
            return InstanceList(v)

        return Boxed(getattr(self.value, key))

    def resolve(self):
        return dict((k,getattr(self.value,k)) for k in self.value.__table__.columns.keys())


class Table(Node):

    def __init__(self, db, value):
        self.db = db
        self.value = value

    def expand(self):
        seq = []

        for instance in self.db.session.query(self.value).all():
             seq.append(Instance(instance))

        return Sequence(seq)

    def resolve(self):
        return self.expand().resolve()


class SQLAlchemy(DataStore):

    def __init__(self, config):
        self.config = config
        self.tables = {}
        self.engine = None
        self._session = None
        self.base = declarative_base()

    def build_column(self, columnspec):
        if "relationship" in columnspec:
            return relationship(columnspec["relationship"])

        types = dict(
            string=String,
            integer=Integer,
            )

        typename = columnspec.get("type", "string")
        try:
            T = types[typename]
        except KeyError:
            self.error("Schema type '%s' is invalid" % typename)

        if "pk" in columnspec and columnspec["pk"]:
            primary_key = True
        else:
            primary_key = False

        args = [T()]

        if "foreign_key" in columnspec:
            args.append(ForeignKey(columnspec['foreign_key']))

        column = Column(*args, primary_key=primary_key)
        return column

    def build_table(self, config):
        if not "name" in config:
            self.error("Table name is not given")

        if not "fields" in config:
            self.error("Field list is missing from definition")

        if not isinstance(config["fields"], list):
            self.error("Field list should be a list of column definitions")

        attrs = {}
        attrs['__tablename__'] = config['name']
        attrs['__table_args__'] = dict(
            useexisting = True,
            )

        for c in config["fields"]:
            attrs[c["name"]] = self.build_column(c)

        table = type(config["name"], (self.base,), attrs)
        return Table(self, table)

    def get_model(self, key):
        model = self.config.get("model").resolve()
        __import__(model)
        m = sys.modules[model]
        for k in dir(m):
            v = getattr(m, k)
            if hasattr(v, "__tablename__") and getattr(v, "__tablename__") == key:
                return Table(self, v)
        #raise NotFound("Table '%s' not defined in '%s'" % (key, model))
        assert False

    def get(self, key):
        if key in self.tables:
            return self.tables[key]

        try:
            if self.config.get("model"):
                tbl = self.tables[key] = self.get_model(key)
                return tbl
        except:
            pass

        for t in self.config.get("tables").resolve():
            if t["name"] == key:
                tbl = self.tables[key] = self.build_table(t)
                return tbl

    @property
    def session(self):
        if self._session:
            return self._session

        if not has_sqlalchemy:
            self.error("You are attempting to use a database from Yay, but SQLAlchemy is not installed")

        if not self.engine:
            self.engine = create_engine(self.config.get("connection").resolve(), echo=True)
        Session = sessionmaker(bind=self.engine)
        self._session = Session()

        return self._session




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node
from yay.errors import NoMatching


class Call(Node):

    def __init__(self, composer, value, params=None):
        self.composer = composer
        self.value = value
        self.params = params
        if params:
            params.set_parent(self)

    def get_context(self, label):
        if self.params:
            try:
                return self.params.expand().get(label)
            except NoMatching:
                pass
        return super(Call, self).get_context(label)

    def get(self, idx):
        return self.expand().get(idx)

    def expand(self):
        try:
            block = self.composer.parent.definitions[self.value]
        except KeyError:
            self.error("'%s' was not defined" % self.value)

        clone = block.clone()
        clone.set_parent(self)

        return clone.expand()

    def resolve(self):
        return self.expand().resolve()

    def walk(self):
        yield self.value

    def clone(self):
        return Call(self.composer, self.value, self.params.clone() if self.params else None)





from yay.nodes import Node, Sequence, Context

class Filter(Node):

    """
    Filter an iterable according to an XPath style selector

    It will be created when the parser encounters something like::

        example: ${foo[@.baz < 5]}

    This will check every object in the sequence ``foo`` for a property
    called ``baz`` with a value less than 5. The output is a sequence.
    """

    def __init__(self, container, filter_expression):
        self.container = container
        if container:
            container.set_parent(self)
        self.filter_expression = filter_expression
        filter_expression.set_parent(self)

    def expand(self):
        resolved = self.container.expand()

        filtered = []
        for r in resolved:
            ctx = Context(self.filter_expression.clone(), {"@": r})
            r.set_parent(ctx)
            ctx.set_parent(self.parent)
            if ctx.resolve():
                filtered.append(r)

        return Sequence(filtered)

    def get(self, idx):
        return self.expand().get(idx)

    def resolve(self):
        return self.expand().resolve()

    def __repr__(self):
        return "Filter(%s, %s)" % (self.container, self.filter_expression)

    def walk(self):
        yield self.container
        yield self.filter_expression

    def clone(self):
        return Filter(self.container.clone(), self.filter_expression.clone())




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, BoxingFactory, Sequence, Context
from yay.nodes.flatten import flatten


class ForEach(Node):

    """
    Represents an iteration over a sequence or other iterable node.

    This will be added to the graph when the parser encounters the ``.foreach`` function::

        example.foreach fruit in basket if fruit.price = 5: ${fruit.name}
    """

    def __init__(self, root, value, args):
        self.root = root
        self.value = value
        value.set_parent(self)

        self.alias = args.pop(0).strip()

        self.lookup = args.pop(0)
        self.lookup.set_parent(self)

        if len(args) and args[0] in ("chain", "nochain", "flatten"):
            self.mode = args.pop(0)
        else:
            self.mode = "flatten"

        if len(args):
            if args[0] != "if":
                self.error("Expected 'if', got '%s'" % args[0])
            args.pop(0)
            self.filter = args.pop(0)
            self.filter.set_parent(self)
        else:
            self.filter = None

    def expand(self):
        lst = []

        for item in self.lookup.expand():
            c = Context(self.value.clone(), {self.alias: item})
            c.set_parent(self)

            if self.filter:
                f = self.filter.clone()
                f.set_parent(c)
                if not f.resolve():
                    continue

            item.set_parent(c)

            if self.mode == "nochain":
                lst.append(c)
            else:
                c.set_parent(self)
                val = c.resolve()
                if isinstance(val, list):
                    if self.mode == "flatten":
                        lst.extend(BoxingFactory.box(v) for v in flatten(val))
                    else:
                        lst.extend(BoxingFactory.box(v) for v in val)
                else:
                    lst.append(BoxingFactory.box(val))

        sq = Sequence(lst)
        sq.set_parent(self.parent)
        return sq

    def resolve(self):
        return self.expand().resolve()

    def walk(self):
        yield self.lookup
        #yield self.value

    def clone(self):
        fe = ForEach(self.root, self.value.clone(), [self.alias, self.lookup.clone(), self.mode])
        if self.filter:
            fe.filter = self.filter.clone()
            fe.filter.set_parent(fe)
        return fe




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib
import urlparse
import StringIO
import os
import sys
import subprocess

from yay.errors import NotFound


class IOpener(object):

    def open(self, uri):
        """ Given a uri, return a YAML compatible stream """
        pass


class FpAdaptor(object):

    def __init__(self, fp):
        self.fp = fp

    def __getattr__(self, key):
        """Everything delegated to the object"""
        return getattr(self.fp, key)


class FileOpener(IOpener):

    schemes = ("file://", )

    def open(self, uri):
        if uri.startswith("file://"):
            uri = uri[8:]
        if not os.path.exists(uri):
            raise NotFound("Local file '%s' could not be found" % uri)

        class File(FpAdaptor):
            @property
            def len(self):
                return int(os.fstat(self.fp.fileno())[6])

        return File(open(uri, "rb"))


class PackageOpener(IOpener):

    schemes = ("package://", )

    def open(self, uri):
        package, uri = uri.lstrip("package://").split("/", 1)
        try:
            __import__(package)
            module = sys.modules[package]
        except ImportError:
            raise NotFound("Package '%s' could not be imported")
        module_path = os.path.dirname(module.__file__)
        path = os.path.join(module_path, uri)
        return FileOpener().open(path)


class HomeOpener(IOpener):

    schemes = ("home://", )

    def open(self, uri):
        uri = os.path.expanduser("~/" + uri.lstrip("home://"))
        return FileOpener().open(uri)


class UrlOpener(IOpener):

    schemes = ("http://", "https://")

    def open(self, uri):
        fp = urllib.urlopen(uri)
        if fp.getcode() != 200:
            raise NotFound("URL '%s' could not be found (HTTP response %s)" % (uri, fp.getcode()))

        class Resource(FpAdaptor):
            @property
            def len(self):
                return int(self.fp.info()['content-length'])

        return Resource(fp)


class MemOpener(IOpener):

    """
    This is purely a testing stub.

    It has a class-shared dictionary called 'data', where keys are uris and
    values are data to return as a StringIO when open() is called. This means
    test cases can exercise the yay extends machinery without the need to create
    temporary files
    """

    schemes = ("mem://", )
    data = {}

    def open(self, uri):
        fp = StringIO.StringIO(self.data[uri])
        fp.len = len(self.data[uri])
        return fp

    @classmethod
    def add(cls, uri, data):
        cls.data[uri] = data

    @classmethod
    def reset(cls):
        cls.data = {}


class Gpg(object):

    def filter(self, fp):
        data = fp.read()
        p = subprocess.Popen(["gpg", "-d"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
        stdout, stderr = p.communicate(data)
        stream = StringIO.StringIO(stdout)
        stream.len = len(stdout)
        stream.secret = True
        return stream


class Openers(object):

    def __init__(self, searchpath=None):
        self.searchpath = searchpath or []

        self.openers = []
        for cls in IOpener.__subclasses__():
            self.openers.append(cls())

    def _scheme(self, uri):
        parsed = urlparse.urlparse(uri)
        return parsed.scheme

    def _absolute(self, uri):
        return self._scheme(uri)

    def _relative(self, uri):
        return not self._absolute(uri)

    def _join(self, *uri):
        if self._scheme(uri[0]):
            retval = uri[0]
            for part in uri[1:]:
                retval = retval.rstrip("/") + "/" + part.lstrip("/")
            return retval
        return os.path.join(*uri)

    def _open(self, uri):
        for opener in self.openers:
            for scheme in opener.schemes:
                if uri.startswith(scheme):
                    return opener.open(uri)
        raise NotFound("Could not find a way to open '%s'" % uri)

    def open(self, uri):
        fp = None

        if uri.startswith("/"):
            fp = FileOpener().open(uri)
        elif self._absolute(uri):
            fp = self._open(uri)
        else:
            for path in self.searchpath:
                try:
                    fp = self._open(self._join(path, uri))
                except NotFound:
                    pass
                else:
                    break
            else:
                fp = FileOpener().open(uri)

        if not fp:
            raise NotFound("'%s' could not be found" % uri)

        if uri.endswith(".gpg"):
            fp = Gpg().filter(fp)

        return fp




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, BoxingFactory

def sum(*args):
    return reduce(lambda x, y: x+y, args)


class Function(Node):

    """
    Created when the parser encounters something that looks like a function call::

        example.foreach i in range(5): ${i}
    """

    functions = {
        "range": range,
        "sum": sum,
        }

    def __init__(self, function, args):
        self.function = function
        self.args = args
        [x.set_parent(self) for x in args]

    def resolve(self):
        args = [arg.resolve() for arg in self.args]
        return self.functions[self.function](*args)

    def expand(self):
        return [BoxingFactory.box(x) for x in self.resolve()]

    def __repr__(self):
        return "Function(%s)" % ", ".join([self.function] + [str(a) for a in self.args])

    def walk(self):
        for arg in self.args:
            yield arg

    def clone(self):
        return Function(self.function, [x.clone() for x in self.args])




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from yay.nodes import Node, BoxingFactory

class Sequence(Node):

    """
    An ordered list of unresolved :py:class:`~yay.nodes.Node` objects.
    """

    def __init__(self, value):
        self.value = value
        [x.set_parent(self) for x in value]

    def get(self, idx):
        return self.value[int(idx)]
        #b = BoxingFactory.box(self.resolve()[int(idx)])
        #return b

    def resolve(self):
        data = []
        for val in self.value:
            data.append(val.resolve())
        return data

    def __iter__(self):
        return iter(self.value)

    def walk(self):
        for val in self.value:
            yield val

    def clone(self):
        return Sequence([x.clone() for x in self.value])


def box_generator(val):
    return Sequence(list(BoxingFactory.box(itm) for itm in val))
BoxingFactory.register(inspect.isgenerator, box_generator)

BoxingFactory.register(lambda x: isinstance(x, list), box_generator)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import yay
from yay.config import Config

from sqlalchemy import Column, String, Integer, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


test_yay = """
injected_data: abc

result: ${injected_data}
"""


class TestInjectPython(unittest.TestCase):

    def test_override_literal(self):
        config = Config()
        config.load(test_yay)
        config.add(dict(injected_data="xyz"))

        self.failUnlessEqual(config.get()["result"], "xyz")




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import yay
from yay.config import Config

from sqlalchemy import Column, ForeignKey, String, Integer, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship

dbyay = """
metadata.bind:
    type: sqlalchemy
    connection: sqlite://
    model: yay.tests.test_datastore_alch_pymodel
"""

Base = declarative_base()

class User(Base):
    __tablename__ = "user"
    id = Column(Integer, primary_key=True)
    username = Column(String)
    password = Column(String)

class Service(Base):
    __tablename__ = "service"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    branch = Column(String)
    host_id = Column(Integer, ForeignKey('host.id'))

class Host(Base):
    __tablename__ = "host"
    id = Column(Integer, primary_key=True)
    name = Column(String)
    services = relationship('Service')


engine = create_engine('sqlite:///:memory:', echo=True)
Session = sessionmaker(bind=engine)
session = Session()


class TestDb(unittest.TestCase):

    def setUp(self):
        self.config = Config()
        self.config.load(dbyay)

        self.t = t = self.config.mapping.get('metadata')

        User = t.get("user").value
        Service = t.get("service").value
        Host = t.get("host").value

        t.backend.engine = engine
        Base.metadata.create_all(engine)

        session.add(User(username='john', password='password'))
        session.add(User(username='john', password='password'))
        session.add(User(username='john', password='password'))
        session.add(User(username='john', password='password'))
        session.add(User(username='john', password='password'))

        h = Host(name="wonderflonium")
        session.add(h)

        s = Service(name="www.foo.com", branch="/trunk")
        h.services.append(s)
        session.add(s)

        session.commit()

    def tearDown(self):
        Base.metadata.drop_all(self.t.backend.engine)

    def test_foreach_host(self):
        self.config.load("""
            test.foreach h in metadata.host: ${h.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ["wonderflonium"])

    def test_foreach_host_service(self):
        self.config.load("""
            test.foreach h in metadata.host:
              .foreach s in h.services: ${s.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ["www.foo.com"])

    def test_foreach_host_service_if(self):
        self.config.load("""
            branch: /trunk
            test.foreach h in metadata.host:
              .foreach s in h.services if s.branch=branch: ${s.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], ["www.foo.com"])

    def test_foreach_host_service_if_failing(self):
        self.config.load("""
            branch: /trun
            test.foreach h in metadata.host:
              .foreach s in h.services if s.branch=branch: ${s.name}
            """)

        self.failUnlessEqual(self.config.get()["test"], [])




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node
from yay import String
from yay.errors import NoMatching


class Comparison(Node):
    def __init__(self, left, right):
        self.left = left
        left.set_parent(self)
        self.right = right
        right.set_parent(self)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.left, self.right)

    def walk(self):
        yield self.left
        yield self.right

    def clone(self):
        return self.__class__(self.left.clone(), self.right.clone())

class And(Comparison):
    def resolve(self):
        return self.left.resolve() and self.right.resolve()

class Or(Comparison):
    def resolve(self):
        return self.left.resolve() or self.right.resolve()

class In(Comparison):
    def resolve(self):
        return self.left.resolve() in self.right.resolve()

class Equal(Comparison):
    def resolve(self):
        return self.left.resolve() == self.right.resolve()

class NotEqual(Comparison):
    def resolve(self):
        return self.left.resolve() != self.right.resolve()

class LessThan(Comparison):
    def resolve(self):
        return self.left.resolve() < self.right.resolve()

class LessThanEqual(Comparison):
    def resolve(self,):
        return self.left.resolve() <= self.right.resolve()

class GreaterThan(Comparison):
    def resolve(self):
        return self.left.resolve() > self.right.resolve()

class GreaterThanEqual(Comparison):
    def resolve(self):
        return self.left.resolve() >= self.right.resolve()


class Access(Node):

    """
    A future lookup on another part of the graph.

    This node is created when the parser encounters any source of lookup.
    For example::

        foo: ${bar.baz.qux}

    This will create 3 Access nodes. The first will look up ``bar`` using
    :py:meth:`.get_context`. The 2nd will lookup ``baz`` on the object returned
    by the ``bar`` lookup. And so on.

    The lookup won't actually happen until :py:meth:`.resolve` is called. This means
    the graph can still continue to change as more data is loaded.
    """

    def __init__(self, container, access):
        self.container = container
        if container:
            container.set_parent(self)
        self.access = access
        self.access.set_parent(self)

    def get(self, idx):
        sr = self.expand()
        return sr.get(idx)

    def expand(self):
        key = self.access.resolve()

        if self.container:
            unresolved = self.container.get(key)
            if unresolved is None:
                self.error("Container does not have field '%s'" % key)
        else:
            unresolved = self.get_context(key)
            if unresolved is None:
                self.error("Context does not have field '%s'" % key)

        if hasattr(unresolved, "expand"):
            return unresolved.expand()

        return unresolved

    def resolve(self):
        sr = self.expand()
        return sr.resolve()

    def walk(self):
        yield self.expand()

    def __repr__(self):
        return "Access(%s, %s)" % (self.container, self.access)

    def clone(self):
        c = None
        if self.container:
            c = self.container.clone()
        return Access(c, self.access.clone())


class Concatenation(Node):

    """
    This node concatenates multiple nodes together. It used used to do
    variable expansion on literals.

    It will be created when the parser encounters something like this:

        foo: The ${animal} jumped over the ${object}
    """

    def __init__(self, *args):
        self.args = args
        [x.set_parent(self) for x in args]

    def resolve(self):
        """
        If any of the nodes that are being concatenated are protected (because
        Yay knows they were loaded from .yay.gpg) then this will return an
        :py:class:`~yay.String` object.

        If none of the nodes  are then it will return a string.
        """
        resolved = [x.resolve() for x in self.args]

        protected = False
        for x in resolved:
            if isinstance(x, String):
                protected = True

        if protected:
            p = String()
            for x in resolved:
                p.add(x)
            return p

        return "".join(str(x) for x in resolved)

    def walk(self):
        for arg in self.args:
            yield arg

    def __repr__(self):
        return "Concat(%s)" % ", ".join(str(arg) for arg in self.args)

    def clone(self):
        return Concatenation(*[x.clone() for x in self.args])


class Else(Node):

    def __init__(self, *children):
        self.children = list(children)
        [c.set_parent(self) for c in children]

    def append(self, child):
        self.children.append(child)
        child.set_parent(self)

    def __repr__(self):
        return "Else()" # % ",".join(self.children)

    def resolve(self):
        for c in self.children:
            try:
                return c.resolve()
            except NoMatching:
                pass
        self.error(NoMatching("No matching field found"))

    def expand(self):
        for c in self.children:
            try:
                c.resolve()
            except NoMatching:
                continue
            else:
                return c.expand()

        self.error(NoMatching("No matching field found"))

    def walk(self):
        for c in self.children:
            yield c

    def clone(self):
        return Else(*[x.clone() for x in self.children])




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, BoxingFactory, Sequence

class Append(Node):

    """
    Appends an iterable to a sequence

    This gets added to the graph when the parser encounters an ``append`` function::

        example.append:
          - foo
          - bar

    If ``example`` hasn't been previously defined it behaves as if it is append
    to an empty sequence.
    """

    def get(self, idx, default=None):
        return BoxingFactory.box(self.resolve()[int(idx)])

    def expand(self):
        if not self.chain:
            return self.value.expand()
        # we initialize this sequence weirdly as we dont want to reparent the nodes we are appending
        s = Sequence([])
        s.value = list(iter(self.chain.expand())) + list(iter(self.value.expand()))
        s.set_parent(self.parent)
        return s

    def resolve(self):
        return self.expand().resolve()

    def walk(self):
        yield self.chain
        yield self.value

    def clone(self):
        a = Append(self.value.clone())
        a.chain = self.chain.clone() if self.chain else None
        return a




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, Boxed, Sequence

def flatten(lst):
    for itm in lst:
        if isinstance(itm, list):
            for x in flatten(itm):
               yield x
        else:
            yield itm

class Flatten(Node):

    """
    Flatten nested lists into a single-dimension

    This node will be added to the graph when the parser encounters the ``.flatten`` function::

        example.flatten:
          - - 1
            - 2
            - 3
          - - a

    It was added to yay to work around :py:class:`yay.nodes.ForEach` returning lists of lists
    when they were nested. This has been resolved, so this node is probably deprecated.
    """

    def expand(self):
        return Sequence(list(Boxed(x) for x in flatten(self.value.resolve())))

    def resolve(self):
        val = self.expand().resolve()
        return val

    def walk(self):
        yield self.value

    def clone(self):
        return Flatten(self.value.clone())




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.errors import EvaluationError


class Node(object):

    """
    The base class for all things that can be inserted into the graph.
    """

    __slots__ = ("chain", "value")
    chain = None
    parent = None
    locked = False

    name = "<Unknown>"
    line = 0
    column = 0
    snippet = None

    def __init__(self, value=None):
        # Premature typing optimisation
        self.value = value
        value.set_parent(self)

    def set_parent(self, parent):
        self.parent = parent

    def resolve(self):
        """
        Resolve an object into a simple type, like a string or a dictionary.

        Node does not provide an implementation of this, all subclasses should
        implemented it.
        """
        raise NotImplementedError(self.resolve)

    def expand(self):
        """
        Generate a simplification of this object that can replace it in the graph
        """
        return self

    def walk(self):
        return iter([])

    def lock(self):
        """
        Attempt to lock a node so any further changes to its value will cause errors.
        """
        self.locked = True
        for child in self.walk():
            child.lock()

    def get_context(self, key):
        """
        Look up value of ``key`` and return it.

        This doesn't do any resolving, the return value will be a subclass of Node.
        """
        if self.parent:
            return self.parent.get_context(key)
        else:
            return self.expand().get(key)

    def get_root(self):
        """
        Find and return the root of this document.
        """
        if self.parent:
            return self.parent.get_root()
        else:
            return self

    def error(self, exc):
        """
        Raise an error message annotated with the line and column of the text that was
        parsed to create this node
        """
        if isinstance(exc, basestring):
            exc = EvaluationError(exc)

        exc.file = self.name            # File
        exc.line = self.line            # Line
        exc.column = self.column        # Column
        exc.snippet = self.snippet      # Snippet

        raise exc

    def clone(self):
        """
        Return a copy of this node.
        """
        raise NotImplementedError(self.clone)

    def __str__(self):
        return repr(self)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect, sys

import yay
from yay.nodes import Node, BoxingFactory, Sequence
from yay.nodes.datastore.bind import DataStore


class InstanceList(Node):

    """
    I am a list of Instance nodes - i.e. a one-to-many relationship
    """

    def __init__(self, value):
        self.value = value

    def get(self, idx):
        v = self.value.all()[idx]
        return BoxingFactory.box(v)

    def resolve(self):
        return list(x.resolve() for x in self.__iter__())

    def __iter__(self):
        for i in range(self.value.count()):
            yield self.get(i)


class Instance(Node):

    def __init__(self, value):
        self.value = value

        self.related = [x.get_accessor_name() for x in self.value._meta.get_all_related_objects()]
        self.many_to_many = [x.get_accessor_name() for x in self.value._meta.get_all_related_many_to_many_objects()]

    def get(self, key):
        v = getattr(self.value, key)

        if key in self.related or key in self.many_to_many:
            return InstanceList(v)

        return BoxingFactory.box(getattr(self.value, key))

    def resolve(self):
        from django.db import models
        mapping = {}
        for k in self.value._meta.get_all_field_names():
            if hasattr(self.value, k) and not isinstance(self.value._meta.get_field_by_name(k)[0], models.fields.related.ForeignKey):
                mapping[k] = self.get(k).resolve()
        return mapping


class Table(Node):

    def __init__(self, value):
        self.value = value

    def get(self, key):
        try:
            idx = int(key)
        except ValueError:
            return BoxingFactory.box(getattr(self.value, key))

        _all = self.value.objects.all()
        return Instance(_all[idx])

    def resolve(self):
        return [Instance(x).resolve() for x in self.value.objects.all()]

    def __iter__(self):
        for i in range(self.value.objects.count()):
            yield self.get(i)


class DjangoStore(DataStore):

    def __init__(self, config):
        self.config = config
        self.tables = {}

    def get_model(self, key):
        model = self.config.get("model").resolve()
        __import__(model)
        m = sys.modules[model]

        if not key in dir(m):
            self.error("Model '%s' not defined in '%s'" % (key, model))

        v = getattr(m, key)

        from django.db import models
        if not inspect.isclass(v) or not issubclass(v, models.Model):
            self.error("'%s' is defined in '%s', but it isn't a Django model" % (key, model))

        return Table(v)

    def get(self, key):
        if key in self.tables:
            return self.tables[key]

        if self.config.get("model"):
            tbl = self.tables[key] = self.get_model(key)
            return tbl

        self.error("'model' not specified")

    def error(self, msg):
        raise RuntimeError(msg)


def lazy_isinstance(classname):
    def _(obj):
        cls = obj.__class__
        if not hasattr(cls, "__mro__"):
            return False
        modname, cname = classname.rsplit(".", 1)
        return any(x.__module__ == modname and x.__name__ == cname for x in cls.__mro__)
    return _

BoxingFactory.register(lazy_isinstance("django.db.models.base.Model"), Instance)
BoxingFactory.register(lazy_isinstance("django.db.models.query.QuerySet"), InstanceList)
BoxingFactory.register(lazy_isinstance("django.db.models.manager.Manager"), InstanceList)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaml.error import MarkedYAMLError
from yaml.events import ScalarEvent, SequenceStartEvent, SequenceEndEvent, \
    MappingStartEvent, MappingEndEvent, AliasEvent, StreamEndEvent

from yay.nodes import *
from yay.parser import Parser
from yay.errors import SyntaxError

class ComposerError(MarkedYAMLError):
    pass


class Composer(object):

    def __init__(self, secret=False):
        self.secret = secret
        self.root = None
        self.parser = Parser(self)

        def define(value, args):
            value.defined_name = args.strip()
            return value

        self.action_map = {
            "define": define,
            "copy": lambda value, args: Copy(value),
            "assign": lambda value, args: value,
            "append": lambda value, args: Append(value),
            "remove": lambda value, args: Remove(value),
            "foreach": lambda value, args: ForEach(self, value, self.parser.foreach_statement.parseString(args)),
            "with": lambda value, args: With(value, *self.parser.as_statement.parseString(args)),
            "select": lambda value, args: Select(value, self.parser.expression.parseString(args)[0]),
            "flatten": lambda value, args: Flatten(value),
            "bind": lambda value, args: Bind(value),
            "secret": lambda value, args: Secret(value),
            }
        self.dirty = False

    def compose(self, previous):
        try:
            return self._compose(previous)
        except MarkedYAMLError, e:
            raise SyntaxError(
                e.problem,
                self.name,
                e.problem_mark.line,
                e.problem_mark.column,
                e.problem_mark.get_snippet()
                )

    def _compose(self, previous):
        # Drop the STREAM-START event.
        self.get_event()

        # Compose a document if the stream is not empty.
        document = None
        if not self.check_event(StreamEndEvent):
            self.get_event() # Drop DOCUMENT-START
            document = self.compose_root_mapping(previous)
            self.get_event() # Drop DOCUMENT-END

        # Ensure that the stream contains no more documents.
        if not self.check_event(StreamEndEvent):
            event = self.get_event()
            raise ComposerError("expected a single document in the stream",
                    document.start_mark, "but found another document",
                    event.start_mark)

        # Drop the STREAM-END event.
        self.get_event()

        return document

    def compose_node(self, previous):
        if self.check_event(AliasEvent):
            raise ComposerError(None, None, "found alias, these arent supported in yay", event.start_mark)

        node = None

        peeked = self.peek_event()

        if self.check_event(ScalarEvent):
            node = self.compose_scalar(previous)
        elif self.check_event(SequenceStartEvent):
            node = self.compose_sequence(previous)
        elif self.check_event(MappingStartEvent):
            node = self.compose_mapping_or_anonymous(previous)

        if not node:
            event = self.peek_event()
            raise ComposerError(None, None, "unexpected event in stream", event.start_mark)

        node.name = self.name
        node.line = peeked.start_mark.line
        node.column = peeked.start_mark.column

        return node

    def compose_scalar(self, previous):
        event = self.get_event()

        if isinstance(event.value, basestring):
            #Icky - this needs to move *beneath* this layer of code
            node = self.parser.templated_string.parseString(event.value)[0]
        else:
            node = self.parser.box(event.value)

        return node

    def compose_sequence(self, previous):
        start = self.get_event()

        data = []
        while not self.check_event(SequenceEndEvent):
            data.append(self.compose_node(None))

        end = self.get_event()

        node = Sequence(data)

        return node

    def handle_imports(self, previous, imports):
        for extend in imports:
            data = self.openers.open(extend)
            secret = hasattr(data, "secret") and data.secret
            previous = self.__class__(data, parent=self.parent, secret=secret).compose(previous)
        return previous

    def handle_special_term(self, previous):
        if self.check_event(MappingEndEvent):
            return previous

        key_event = self.peek_event()
        if key_event.value != self.special_term:
            return previous
        self.get_event()

        special_term = self.compose_node(None).resolve()

        return self.handle_imports(previous, special_term.get("extends", []))

    def compose_mapping_or_anonymous(self, previous):
        start = self.get_event()

        try:
            if not self.check_event(MappingEndEvent):
                if self.peek_event().value.startswith("."):
                    return self.compose_anonymous(previous)
                elif self.peek_event().value.endswith("!"):
                    n = Call(self, self.get_event().value[:-1], self.compose_node(None))
                    return n
            return self.compose_mapping(previous)
        finally:
            self.get_event()

    def compose_anonymous(self, previous):
        # An anonymous expression that doesnt bind to a keyword
        key_event = self.get_event()
        action = key_event.value[1:]

        action_args = None
        if " " in action:
            action, action_args = action.split(" ", 1)

        value = self.compose_node(None)

        assert self.check_event(MappingEndEvent)

        return self.action_map[action](value, action_args)

    def compose_mapping_value(self, container):
        key_event = self.get_event()
        key = key_event.value

        action = "assign"
        if "." in key and key != '.include':
            key, action = key.split(".", 1)

        action_args = None
        if " " in action:
            action, action_args = action.split(" ", 1)

        if action == "define":
            key = ".define"

        try:
            existing = container.get(key)
        except:
            existing = None

        # Grab scalar value
        boxed = self.compose_node(existing)

        # Further box the value based on the kind of action it is
        boxed = self.action_map[action](boxed, action_args)

        # Can't override a locked node
        if existing and existing.locked:
            boxed.error("%s is locked and cannot be overidden" % key)

        # Make sure that Appends are hooked up to correct List
        boxed.chain = existing

        return key, boxed

    def compose_mapping(self, previous):
        container = Mapping(previous)
        while not self.check_event(MappingEndEvent):
            key, value = self.compose_mapping_value(container)
            container.set(key, value)
        return container

    def compose_root_mapping(self, previous):
        if not self.check_event(MappingStartEvent):
            ev = self.get_event()
            #raise ComposerError("Expected root mapping - am denied", ev.start_mark)
            return previous

        start = self.get_event()

        previous = self.handle_special_term(previous)

        previous = Mapping(previous)
        while not self.check_event(MappingEndEvent):
            key, value = self.compose_mapping_value(previous)
            if key == ".include":
                value.set_parent(previous)

                if isinstance(value, Sequence):
                    for i in value.value:
                        i.lock()
                        include = i.resolve()
                        previous = self.handle_imports(previous, [include])
                else:
                    value.lock()
                    includes = value.resolve()

                    if not isinstance(includes, list):
                        value.error("Expected something that resolved to a sequence and didnt")

                    previous = self.handle_imports(previous, includes)

                previous = Mapping(previous)

            elif key == ".define":
                self.parent.definitions[value.defined_name] = value

            else:
                previous.set(key, value)

        end = self.get_event()

        return previous




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from yay.loader import Loader
from yay.openers import Openers
from yay.nodes import BoxingFactory, Mapping
from yay.errors import ProgrammingError


class Config(object):

    def __init__(self, special_term='yay', searchpath=None):
        self.special_term = special_term
        self.openers = Openers(searchpath=searchpath)
        self.clear()
        self.mapping = None

    def load_uri(self, uri):
        stream = self.openers.open(uri)
        self.load(stream, uri, hasattr(stream, "secret") and stream.secret)

    def load(self, stream, name="<Unknown>", secret=False):
        l = Loader(stream, name=name, parent=self, secret=secret)
        data = l.compose(self.mapping)
        self.mapping = data

    def add(self, data):
        boxed = BoxingFactory.box(data)
        if not isinstance(boxed, Mapping):
            raise ProgrammingError("Tried to call Config.add on type '%s'. This cannot be boxed as a 'Mapping'." % type(data))

        boxed.predecessor = self.mapping
        if boxed.predecessor:
            boxed.predecessor.set_parent(boxed)

        self.mapping = boxed

    def clear(self):
        self.mapping = None
        self.definitions = {}

    def get(self):
        if not self.mapping:
            return {}
        return self.mapping.resolve()

def load_uri(uri, special_term='yay'):
    c = Config(special_term)
    c.load_uri(uri)
    return c.get()

def load(stream, special_term='yay', secret=False):
    c = Config(special_term)
    c.load(stream, secret)
    return c.get()

def dump(obj):
    return yaml.dump(obj, default_flow_style=False)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node

class Context(Node):

    """
    Add context to the graph.

    Any children of this node will get different context to the parents
    of this node as it implements ``get_context``.
    """

    def __init__(self, value, context):
        super(Context, self).__init__(value)
        self.context = context

    def get_context(self, key):
        """
        If ``key`` is provided by this node return it, otherwise fall
        back to default implementation.
        """
        val = self.context.get(key, None)
        if not val:
            val = super(Context, self).get_context(key)
        return val

    def expand(self):
        return self.value.expand()

    def resolve(self):
        return self.value.resolve()

    def __repr__(self):
        return "Context(%s)" % self.value

    def clone(self):
        return Context(self.value.clone(), self.context)

